/**!
 * gdk-pixbuf.notes:
 *
 * Functions with a GTimeVal* argument will take a integer with is treated
 * as usecs.
 **/


/**!
 * gdk-pixbuf.func.gdk::pixbuf_animation new_from_file:
 * gdk::pixbuf_animation new_from_file
 * gdk_pixbuf_animation_new_from_file
 *
 *> Usage
 *   @name          %char*
 *   @name          %GError**
 *   Return value   %GdkPixbufAnimation*
 *
 * Wraps %gdk_pixbuf_animation_new_from_file.
 **/

static int cmd_gdk_pixbuf_animation_new_from_file (Tcl_Interp *interp,
						   int objc,
						   Tcl_Obj *CONST objv[],
						   const cmd_sig_t *sig)
{
  cval_t *arg;
  GError *err = 0;

  /* G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 4, sig, 0X2, "GErrorVar");

  if (!arg) {
    return TCL_ERROR;
  }

  /* GdkPixbufAnimation* {const char*} GError** */
  arg[0].ptr_val = gdk_pixbuf_animation_new_from_file (arg[1].ptr_val, &err);

  if (!arg[0].ptr_val) {
    if (setGErrorVar (interp, objv[3], err) != TCL_OK) {
      cval_free (arg, sig->type, sig->len);
      return TCL_ERROR;
    }
  }

  /* GDK_TYPE_PIXBUF_ANIMATION */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gdk-pixbuf.func.gdk::pixbuf new_from_file:
 * gdk::pixbuf new_from_file
 * gdk_pixbuf_new_from_file
 *
 *> Usage
 *   @name         %char*
 *   @name         %GError**
 *   Return value  %GdkPixbuf*
 *
 * Wraps %gdk_pixbuf_new_from_file.
 **/

static int cmd_gdk_pixbuf_new_from_file (Tcl_Interp *interp, int objc,
				  Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GError *err = 0;
  
  /* G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 4, sig, 0X2, "GErrorVar");
  
  if (!arg) {
    return TCL_ERROR;
  }
  
  /* GdkPixbuf* {const char*} GError** */
  arg[0].ptr_val = gdk_pixbuf_new_from_file (arg[1].ptr_val, &err);
  
  if (!arg[0].ptr_val) {
    if (setGErrorVar (interp, objv[3], err) != TCL_OK) {
      cval_free (arg, sig->type, sig->len);
      return TCL_ERROR;
    }
  }

  /* GDK_TYPE_PIXBUF */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));
  
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gdk-pixbuf.func.gdk::pixbuf get_pixels:
 * gdk_pixbuf_get_pixels
 * gdk::pixbuf get_pixels
 *
 *> Usage
 *   @obj:          %GdkPixbuf*
 *   Return value   %guchar*
 *
 * Wraps %gdk_pixbuf_get_pixels. Returns a copy of all the pixel data
 * in @obj as binary data.
 **/

static int cmd_gdk_pixbuf_get_pixels (Tcl_Interp *interp, int objc,
			       Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  int len;

  /* GDK_TYPE_PIXBUF */
  arg = cval_gather (interp, objc, objv, 3, sig, 0);

  if (!arg) {
    return TCL_ERROR;
  }

  /* guchar* {const GdkPixbuf*} */
  arg[0].ptr_val = gdk_pixbuf_get_pixels (arg[1].ptr_val);

  len = gdk_pixbuf_get_height (arg[1].ptr_val) *
    gdk_pixbuf_get_rowstride (arg[1].ptr_val);
  
  Tcl_SetObjResult (interp, Tcl_NewByteArrayObj (arg[0].ptr_val, len));
		    
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gdk-pixbuf.func.gdk::pixbuf new_from_data:
 * gdk::pixbuf new_from_data
 * gdk_pixbuf_new_from_data
 *
 *> Usage
 *   @data          %guchar*
 *   @name:         %GdkColorspace
 *   @name:         %gboolean
 *   @name:         %int
 *   @name:         %int
 *   @name:         %int
 *   @name:         %int
 *   Return value   %GdkPixbuf*
 *
 * Wraps %gdk_pixbuf_new_from_data. No destroy callback since our
 * implementation needs it.
 **/

static void gdk_pixbuf_destroyed (guchar *pixels, gpointer data)
{
  Tcl_Obj *obj = data;

  Tcl_DecrRefCount (obj);
}

static int cmd_gdk_pixbuf_new_from_data (Tcl_Interp *interp, int objc, 
				  Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  guchar *data;
  Tcl_Obj *obj;

  /* GDK_TYPE_COLORSPACE G_TYPE_BOOLEAN G_TYPE_INT G_TYPE_INT G_TYPE_INT
   * G_TYPE_INT
   */
  arg = cval_gather (interp, objc, objv, 9, sig, 0X1, "bytes");

  if (!arg) {
    return TCL_ERROR;
  }

  /* pixbuf owns the memory, so we have to protect the Tcl_Obj */

  obj = objv[2];

  if (Tcl_IsShared (obj)) {
    obj = Tcl_DuplicateObj (obj);
  }

  Tcl_IncrRefCount (obj);
  data  = Tcl_GetByteArrayFromObj (objv[2], 0);

  /* GdkPixbuf* {const guchar*} GdkColorspace gboolean int int int int
   * GdkPixbufDestroyNotify gpointer
   */
  arg[0].ptr_val = gdk_pixbuf_new_from_data (data, arg[1].int_val,
					     arg[2].int_val, arg[3].int_val,
					     arg[4].int_val, arg[5].int_val,
					     arg[6].int_val,
					     gdk_pixbuf_destroyed, obj);

  /* GDK_TYPE_PIXBUF */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gdk-pixbuf.func.gdk::pixbuf new_from_xpm_data:
 * gdk::pixbuf new_from_xpm_data
 * gdk_pixbuf_new_from_xpm_data
 *
 *> Usage
 *   @data          {string ...}
 *   Return value   %GdkPixbuf*
 *
 * Wraps %gdk_pixbuf_new_from_xpm_data. @data is a list of the strings
 * making up the xpm.
 **/

static int cmd_gdk_pixbuf_new_from_xpm_data (Tcl_Interp *interp, int objc,
					     Tcl_Obj *CONST objv[],
					     const cmd_sig_t *sig)
{
  cval_t *arg;
  Tcl_Obj **objv_data;
  const char **str_data;

   arg = cval_gather (interp, objc, objv, 3, sig, 0X1, "{str}");

  if (!arg) {
    return TCL_ERROR;
  }

  if (Tcl_ListObjGetElements (interp, objv[2], &objc, &objv_data) != TCL_OK) {
    cval_free (arg, sig->type, sig->len);
    return TCL_ERROR;
  }

  str_data = g_new (const char *, objc);
  while (objc--) {
    str_data[objc] = Tcl_GetString (objv_data[objc]);
  }

  /* GdkPixbuf* {const char**} */
  arg[0].ptr_val = gdk_pixbuf_new_from_xpm_data (str_data);

  /* GDK_TYPE_PIXBUF */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  g_free (str_data);
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gdk-pixbuf.func.gdk::pixbuf save:
 * gdk::pixbuf save
 * gdk_pixbuf_savev
 *
 *> Usage
 *   @name          %GdkPixbuf*
 *   @name          %char*
 *   @name          %char*
 *   @options       {option value ....}
 *   @name          %GError**
 *   Return value   %gboolean
 *
 * Wraps %gdk_pixbuf_savev. @options is a list of keys and values. 
 **/

static int cmd_gdk_pixbuf_save (Tcl_Interp *interp, int objc,
				Tcl_Obj *CONST objv[],
				const cmd_sig_t *sig)
{
  cval_t *arg;
  GError *err = 0;
  char **opt, **val;
  int i, len;
  Tcl_Obj **obj;

  /* GDK_TYPE_PIXBUF G_TYPE_STRING G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 7, sig, 0X18, "{opt val ...}",
		     "GErrorVar");

  if (!arg) {
    return TCL_ERROR;
  }

 if (Tcl_ListObjGetElements(interp, objv[5], &len, &obj) != TCL_OK) {
    return TCL_ERROR;
  }

  if (len & 1) {
    appendResult (interp, "no value for an option", (char *) 0);
    return TCL_ERROR;
  }

  len /= 2;
  opt = g_new (char *, len + 1);
  val = g_new (char *, len + 1);

  for (i = 0; i < len; i++) {
    opt[i] = Tcl_GetString (*(obj++));
    val[i] = Tcl_GetString (*(obj++));
  }
  opt[i] = 0;
  val[i] = 0;

  /* gboolean GdkPixbuf* {const char*} {const char*} char** char** GError** */
  arg[0].int_val = gdk_pixbuf_savev (arg[1].ptr_val, arg[2].ptr_val,
				     arg[3].ptr_val, opt, val, &err);

  if (!arg[0].int_val) {
    if (setGErrorVar (interp, objv[6], err) != TCL_OK) {
      g_free (opt);
      g_free (val);
      cval_free (arg, sig->type, sig->len);
      return TCL_ERROR;
    }
  }

  /* G_TYPE_BOOLEAN */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  g_free (opt);
  g_free (val);
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gdk-pixbuf.func.gdk::pixbuf_loader new_with_type:
 * gdk_pixbuf_loader_new_with_type
 * gdk::pixbuf_loader new_with_type
 *
 *> Usage
 *   @name         %char*
 *   @name         %GError**
 *   Return value  %GdkPixbufLoader*
 *
 * Wraps %gdk_pixbuf_loader_new_with_type.
 **/

static int cmd_gdk_pixbuf_loader_new_with_type (Tcl_Interp *interp, int objc,
					 Tcl_Obj *CONST objv[],
					 const cmd_sig_t *sig)
{
  cval_t *arg;
  GError *err = 0;

  /* G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 4, sig, 0X2, "GErrorVar");

  if (!arg) {
    return TCL_ERROR;
  }

  /* GdkPixbufLoader* {const char*} GError** */
  arg[0].ptr_val = gdk_pixbuf_loader_new_with_type (arg[1].ptr_val, &err);

  if (!arg[0].ptr_val) {
    if (setGErrorVar (interp, objv[3], err) != TCL_OK) {
      cval_free (arg, sig->type, sig->len);
      return TCL_ERROR;
    }
  }

  /* GDK_TYPE_PIXBUF_LOADER */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gdk-pixbuf.func.gdk::pixbuf_loader write:
 * gdk::pixbuf_loader write
 * gdk_pixbuf_loader_write
 *
 *> Usage
 *   @name         %GdkPixbufLoader*
 *   @name         %guchar*
 *   @name         %gsize
 *   @name         %GError**
 *   Return value  %gboolean
 *
 * Wraps %gdk_pixbuf_loader_write.
 **/

static int cmd_gdk_pixbuf_loader_write (Tcl_Interp *interp, int objc,
				 Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GError *err = 0;
  guchar *data;

  /* GDK_TYPE_PIXBUF_LOADER G_TYPE_UINT */
  arg = cval_gather (interp, objc, objv, 5, sig, 0X6, "bytes", "GErrorVar");

  if (!arg) {
    return TCL_ERROR;
  }

  data = Tcl_GetByteArrayFromObj (objv[3], &objc);

  /* gboolean GdkPixbufLoader* {const guchar*} gsize GError** */
  arg[0].int_val = gdk_pixbuf_loader_write (arg[1].ptr_val, data,
					    objc, &err);

  if (!arg[0].int_val) {
    if (setGErrorVar (interp, objv[3], err) != TCL_OK) {
      cval_free (arg, sig->type, sig->len);
      return TCL_ERROR;
    }
  }

  /* G_TYPE_BOOLEAN */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gdk-pixbuf.func.gdk::pixbuf_loader close:
 * gdk::pixbuf_loader close
 * gdk_pixbuf_loader_close
 *
 *> Usage
 *   @name         %GdkPixbufLoader*
 *   @name         %GError**
 *   Return value  %gboolean
 *
 * Wraps %gdk_pixbuf_loader_close.
 **/

static int cmd_gdk_pixbuf_loader_close (Tcl_Interp *interp, int objc,
				 Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GError *err = 0;

  /* GDK_TYPE_PIXBUF_LOADER */
  arg = cval_gather (interp, objc, objv, 4, sig, 0X2, "GErrorVar");

  if (!arg) {
    return TCL_ERROR;
  }

  /* gboolean GdkPixbufLoader* GError** */
  arg[0].int_val = gdk_pixbuf_loader_close (arg[1].ptr_val, &err);

  if (!arg[0].int_val) {
    if (setGErrorVar (interp, objv[3], err) != TCL_OK) {
      cval_free (arg, sig->type, sig->len);
      return TCL_ERROR;
    }
  }

  /* G_TYPE_BOOLEAN */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gdk-pixbuf.func.gdk::pixbuf_animation_iter advance:
 * gdk::pixbuf_animation_iter advance
 * gdk_pixbuf_animation_iter_advance
 * 
 *> Usage
 *   @name         %GdkPixbufAnimationIter*
 *   @name         %GTimeVal*
 *   Return value  %gboolean
 *
 * Wraps %gdk_pixbuf_animation_iter_advance.
 **/

static int cmd_gdk_pixbuf_animation_iter_advance (Tcl_Interp *interp, int objc,
					   Tcl_Obj *CONST objv[],
					   const cmd_sig_t *sig)
{
  cval_t *arg;
  GTimeVal t;

  /* GDK_TYPE_PIXBUF_ANIMATION_ITER */
  arg = cval_gather (interp, objc, objv, 4, sig, 0X2, "usec");

  if (!arg) {
    return TCL_ERROR;
  }

  t.tv_sec = 0;

  if (Tcl_GetLongFromObj (interp, objv[3], &t.tv_usec) != TCL_OK) {
    return TCL_ERROR;
  }

  /* gboolean GdkPixbufAnimationIter* {const GTimeVal*} */
  arg[0].int_val = gdk_pixbuf_animation_iter_advance (arg[1].ptr_val, &t);

  /* G_TYPE_BOOLEAN */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gdk-pixbuf.func.gdk::pixbuf_animation get_iter:
 * gdk::pixbuf_animation get_iter
 * gdk_pixbuf_animation_get_iter
 *
 *> Usage
 *   @name:         %GdkPixbufAnimation*
 *   @name:         %GTimeVal*
 *   Return value   %GdkPixbufAnimationIter*
 *
 * Wraps %gdk_pixbuf_animation_get_iter.
 **/

static int cmd_gdk_pixbuf_animation_get_iter (Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GTimeVal t;

  /* GDK_TYPE_PIXBUF_ANIMATION */
  arg = cval_gather (interp, objc, objv, 4, sig, 0X2, "usec");

  if (!arg) {
    return TCL_ERROR;
  }

  t.tv_sec = 0;

  if (Tcl_GetLongFromObj (interp, objv[3], &t.tv_usec) != TCL_OK) {
    return TCL_ERROR;
  }

  /* GdkPixbufAnimationIter* GdkPixbufAnimation* {const GTimeVal*} */
  arg[0].ptr_val = gdk_pixbuf_animation_get_iter (arg[1].ptr_val, &t);

  /* GDK_TYPE_PIXBUF_ANIMATION_ITER */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}
