/**!
 * gnome.notes:
 *
 * I wrapped the needed initialization functions and widgets that looked
 * useful. Much of the old functionality is now available in Gtk 2.0 and the
 * gnome documentation hasn't been updated in a while which made choosing
 * quite difficult.
 *
 * Gnome programs must call 'gnome::program init' before doing anything else.
 **/

/**!
 * gnome.func.gnome::program init:
 * gnome::program init
 * gnome_program_init
 *
 *> Usage
 *   @name       program name
 *   @version    program version
 *   @argv       list of arguments to pass in
 *
 * Wraps %gnome_program_init.
 **/

static int cmd_gnome_program_init (Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  int i, argc;
  char **argv;
  Tcl_Obj **obj;

  /* G_TYPE_STRING G_TYPE_STRING GNOME_TYPE_MODULE_INFO G_TYPE_INT G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 5, sig, 0x4, "{argv}");

  if (!arg) {
    return TCL_ERROR;
  }

  if (Tcl_ListObjGetElements (interp, objv[4], &argc, &obj) != TCL_OK) {
    return TCL_ERROR;
  }

  argc++;
  argv = g_new (char *, argc);

  /* gnome_init expects a C like argv ??*/

  argv[0] = (char *) Tcl_GetNameOfExecutable ();
    
  g_assert (argv[0] != 0);
  
  for (i = 1; i < argc; i++) {
    argv[i] = Tcl_GetString (objv[i - 1]);
  }

  /* GnomeProgram* {const char*} {const char*} {const GnomeModuleInfo*} int char** {const char*} ... */
  arg[0].ptr_val = gnome_program_init (arg[1].ptr_val, arg[2].ptr_val,
				        LIBGNOMEUI_MODULE, argc, argv, 0, 0);

  /* GNOME_TYPE_PROGRAM */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  g_free (argv);
  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gnome.func.gnome::about new:
 * gnome::about new
 * gnome_help_display
 *
 *> Usage
 *   @name:         %char*
 *   @version:      %char*
 *   @copyright:    %char*
 *   @comments:     %char*
 *   @authors:      list of names
 *   @documenters:  list of names
 *   @translator:   %char*
 *   @logo:         %GdkPixbuf*
 *   Return value   %GtkWidget%
 *
 * Wraps %gnome_help_display.
 **/

static int cmd_gnome_about_new (Tcl_Interp *interp, int objc,
				Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  char **auth, **doc;

  /* G_TYPE_STRING G_TYPE_STRING G_TYPE_STRING G_TYPE_STRING G_TYPE_STRING GDK_TYPE_PIXBUF */
  arg = cval_gather (interp, objc, objv, 10, sig, 0X30, "const gchar**", "const gchar**");

  if (!arg) {
    return TCL_ERROR;
  }

  if (getArgv (interp, objv[6], &auth) != TCL_OK) {
    return TCL_ERROR;
  }

  if (getArgv (interp, objv[7], &doc) != TCL_OK) {
    return TCL_ERROR;
  }
  

  /* GtkWidget* {const gchar*} {const gchar*} {const gchar*} {const gchar*} {const gchar**} {const gchar**} {const gchar*} GdkPixbuf* */
  arg[0].ptr_val = gnome_about_new (arg[1].ptr_val, arg[2].ptr_val,
				    arg[3].ptr_val, arg[4].ptr_val,
				    (const gchar **) auth,
				    (const gchar **) doc,
				    arg[5].ptr_val, arg[6].ptr_val);

  /* GTK_TYPE_WIDGET */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}


/**!
 * gnome.func.gnome::help display:
 * gnome::help display
 * gnome_help_display
 *
 *> Usage
 *   @name:         %char*
 *   @name:         %char*
 *   @name:         %GErrorVar
 *   Return value   %gboolean
 *
 * Wraps %gnome_help_display.
 **/

static int cmd_gnome_help_display (Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GError *err = 0;

  /* G_TYPE_STRING G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 5, sig, 0X4, "GErrorVar");

  if (!arg) {
    return TCL_ERROR;
  }

  /* gboolean {const char*} {const char*} GError** */
  arg[0].int_val = gnome_help_display (arg[1].ptr_val, arg[2].ptr_val, &err);

  if (!arg[0].int_val) {
    if (setGErrorVar (interp, objv[4], err) != TCL_OK) {
      cval_free (arg, sig->type, sig->len);
      return TCL_ERROR;
    }
  }

  /* G_TYPE_BOOLEAN */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}

/**!
 * gnome.func.gnome::url show:
 * gnome::url show
 * gnome_url_show
 *
 *> Usage
 *   @name          %char*
 *   @name          %GErrorVar
 *   Return value   %gboolean
 *
 * Wraps %gnome_url_show.
 **/

static int cmd_gnome_url_show (Tcl_Interp *interp, int objc,
			       Tcl_Obj *CONST objv[], const cmd_sig_t *sig)
{
  cval_t *arg;
  GError *err = 0;

  /* G_TYPE_STRING */
  arg = cval_gather (interp, objc, objv, 4, sig, 0X2, "GErrorVar");

  if (!arg) {
    return TCL_ERROR;
  }

  /* gboolean {const char*} GError** */
  arg[0].int_val = gnome_url_show (arg[1].ptr_val, 0);

  if (!arg[0].int_val) {
    if (setGErrorVar (interp, objv[3], err) != TCL_OK) {
      cval_free (arg, sig->type, sig->len);
      return TCL_ERROR;
    }
  }

  /* G_TYPE_BOOLEAN */
  Tcl_SetObjResult (interp, newType (sig->type[0], &arg[0]));

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}
