#include <tcl-gobject.h>

/* #define TRACE 1 */ /* if defined, print out function calls as they occur */

static GQuark cmd_key;

static int get_bool (Tcl_Interp *interp, Tcl_Obj *obj, cval_t *v, GType type)
{
  return Tcl_GetBooleanFromObj (interp, obj, &v->int_val);
}

static Tcl_Obj *new_bool (cval_t *v, GType type)
{
  return Tcl_NewBooleanObj (v->int_val);
}

static int get_double (Tcl_Interp *interp, Tcl_Obj *obj, cval_t *v,
		      GType type)
{        
  return Tcl_GetDoubleFromObj (interp, obj, &v->double_val);
}

static Tcl_Obj *new_double (cval_t *v, GType type)
{
  return Tcl_NewDoubleObj (v->double_val);
}

static int get_float (Tcl_Interp *interp, Tcl_Obj *obj, cval_t *v,
		      GType type)
{
  int code = Tcl_GetDoubleFromObj (interp, obj, &v->double_val);
  v->float_val = v->double_val;
  return code;
}

static Tcl_Obj *new_float (cval_t *v, GType type)
{
  return Tcl_NewDoubleObj (v->float_val);
}

static int get_int (Tcl_Interp *interp, Tcl_Obj *obj, cval_t *v,
		    GType type)
{
  return Tcl_GetIntFromObj (interp, obj, &v->int_val);
}

static Tcl_Obj *new_int ( cval_t *v, GType type)
{
  return Tcl_NewIntObj (v->int_val);
}

static int get_char (Tcl_Interp *interp, Tcl_Obj *obj, cval_t *v,
		    GType type)
{
  int code = Tcl_GetIntFromObj (interp, obj, &v->int_val);
  v->char_val = v->int_val;
  return code;
}

static Tcl_Obj *new_char (cval_t *v, GType type)
{
  return Tcl_NewIntObj (v->char_val);
}

static int get_long (Tcl_Interp *interp, Tcl_Obj *obj, cval_t *v,
		    GType type)
{
  return Tcl_GetLongFromObj (interp, obj, &v->long_val);
}

static Tcl_Obj *new_long (cval_t *v, GType type)
{
  return Tcl_NewLongObj (v->long_val);
}

static int get_handle (Tcl_Interp *interp, Tcl_Obj *obj, cval_t *v, GType type)
{
  return getHandle (interp, obj, type, &v->ptr_val);
}

static Tcl_Obj *new_handle (cval_t *v, GType type)
{
  return newHandle (v->ptr_val, type);
}

static int get_enum (Tcl_Interp *interp, Tcl_Obj *obj, cval_t *v,
		    GType type)
{
  return getGEnum (interp, obj, type, &v->int_val);
}

static Tcl_Obj *new_enum (cval_t *v, GType type)
{
  return newGEnum (type, v->int_val);
}

static int get_flags (Tcl_Interp *interp, Tcl_Obj *obj, cval_t *v,
		    GType type)
{
  return getGFlags (interp, obj, type, &v->int_val);
}

static Tcl_Obj *new_flags (cval_t *v, GType type)
{
  return newGFlags (type, v->int_val);
}

static int get_string (Tcl_Interp *interp, Tcl_Obj *obj, cval_t *v, GType type)
{
  v->ptr_val = Tcl_GetString (obj);
  return TCL_OK;
}

static Tcl_Obj *new_string (cval_t *v, GType type)
{
  return Tcl_NewStringObj (v->ptr_val, -1);
}

static int get_boxed (Tcl_Interp *interp, Tcl_Obj *obj, cval_t *v, GType type)
{
  if (type == G_TYPE_CLOSURE) {
    v->ptr_val = getGClosure (interp, obj);
    return TCL_OK;
  } else {
    return getHandle (interp, obj, type, &v->ptr_val);
  }
}

static Tcl_Obj *new_boxed (cval_t *v, GType type)
{
  return newHandle (v->ptr_val, type);
}

typedef int (*cget_func) (Tcl_Interp *, Tcl_Obj *, cval_t *, GType type);
typedef Tcl_Obj *(*cnew_func) (cval_t *, GType);

static const struct {
  const cget_func cget;
  const cnew_func cnew;
} map[] = {
  {0, 0},                     /* invalid */
  {0, 0},                     /* none */
  {get_handle, new_handle},   /* interface */
  {get_char, new_char},       /* char */
  {get_char, new_char},       /* uchar */
  {get_bool, new_bool},       /* bool */
  {get_int, new_int},         /* int */
  {get_int, new_int},         /* uint */
  {get_long, new_long},       /* long */
  {get_long, new_long},       /* ulong */
  {get_long, new_long},       /* int64 */
  {get_long, new_long},       /* uint64 */
  {get_enum, new_enum},       /* GEnum */
  {get_flags, new_flags},     /* GFlags */
  {get_float, new_float},     /* float */
  {get_double, new_double},   /* double */
  {get_string , new_string},  /* string */
  {get_handle, new_handle},   /* pointer */
  {get_boxed, new_boxed},     /* boxed */
  {0, 0},                     /* param */
  {get_handle, new_handle},   /* GObject */
};

int getType (Tcl_Interp *interp, Tcl_Obj *obj, GType type, cval_t *v)
{
  GType t = G_TYPE_FUNDAMENTAL (type) >> G_TYPE_FUNDAMENTAL_SHIFT;

  g_assert (t < G_N_ELEMENTS (map));
  g_assert (map[t].cget != 0);

  return map[t].cget (interp, obj, v, type);
}

Tcl_Obj *newType (GType type, cval_t *v)
{
  GType t = G_TYPE_FUNDAMENTAL (type) >> G_TYPE_FUNDAMENTAL_SHIFT;

  g_assert (t < G_N_ELEMENTS (map));
  g_assert (map[t].cnew != 0);

  return map[t].cnew (v, type);
}

void cval_free (cval_t *arg, const GType *type, int len) 
{
  GType t;

  while (len--) {
    t = G_TYPE_FUNDAMENTAL (type[len]);

    if (t == G_TYPE_STRING_DYNAMIC) {
      g_free (arg[len].ptr_val);
    }
  }

  g_free (arg);
}

void cmd_register (Tcl_Interp *interp, GType type, const cmd_t *cmd,
		   char *proc)
{
  if (!cmd_key) {
    cmd_key =  g_quark_from_static_string ("_cmd_t");
  }

  if (type != G_TYPE_INVALID) {
    g_type_set_qdata (type, cmd_key, (gpointer) cmd);
  }

  Tcl_CreateObjCommand (interp, proc, cmd_eval, (ClientData) cmd, 0);
}

static void wrong_num_args (Tcl_Interp *interp, Tcl_Obj *CONST objv[],
			    const GType *type, int len, int n)
{
  int i;
  GString *usage = g_string_sized_new (16);

  for (i = 1; i < len; i++) {
    g_string_append (usage, g_type_name (type[i]));
    g_string_append_c (usage, ' ');
  }

  g_string_truncate (usage, usage->len - 1);

  Tcl_WrongNumArgs (interp, n, objv, usage->str);

  g_string_free (usage, TRUE);
}

static void cmd_print(const cmd_sig_t *sig, int objc, Tcl_Obj *CONST objv[])
{
  int i;
  GType type;

  printf ("Tcl command: ");
  
  for (i = 0; i < objc; i++) {
    printf("'%s' ", Tcl_GetString(objv[i]));
  }

  printf ("Types: \n");

  for (i = 1; i < sig->len; i++) {
    type = sig->type[i];
    printf ("\t%d %s {%s}\n", type, g_type_name (type),
	    g_type_name (g_type_fundamental (type)));
  }

  type = sig->type[0];
  printf ("\tReturns: %d %s\n", type, g_type_name (type));
}

static int do_cmd_eval (const cmd_t *cmd, Tcl_Interp *interp, int objc,
			Tcl_Obj *CONST objv[])
{
  int i, len;
  const GType *type;
  cval_t *arg;


#ifdef TRACE
  cmd_print (cmd->sig, objc, objv);
#endif

  if (!cmd->sig->wrap) {
    return ((eval_func) cmd->func) (interp, objc, objv, cmd->sig);
  }

  type = cmd->sig->type;
  len = cmd->sig->len;

  if (len + 1 != objc) {
    wrong_num_args (interp, objv, type, len, 2);
    return TCL_ERROR;
  }

  arg = g_new (cval_t, len);  
  objv++;

  for  (i = 1; i < len; i++) {
    if (getType (interp, objv[i], type[i], &arg[i]) != TCL_OK) {
      cval_free (arg, type, i);
      return TCL_ERROR;
    }
  }

  cmd->sig->wrap (cmd->func, arg);

  if (type[0] != G_TYPE_NONE) {
    Tcl_SetObjResult (interp, newType (type[0], &arg[0]));
  }
  
  cval_free (arg, type, len);
  return TCL_OK;  
}

int cmd_eval (ClientData clientData, Tcl_Interp *interp, int objc,
	     Tcl_Obj *CONST objv[])
{
  const cmd_t *cmd = clientData;
  int i;

  if (objc < 2) {
    Tcl_WrongNumArgs (interp, 1, objv, "option arg...");
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObjStruct (interp, objv[1], (char **) &cmd[0].name,
				 sizeof (cmd_t), "cmd", TCL_EXACT, &i)
      != TCL_OK) {
    return TCL_ERROR;
  }

  cmd += i;

  return do_cmd_eval (cmd, interp, objc, objv);
}

/* FIXME: handle interfaces? */

int cmd_invoke (Tcl_Interp *interp, GType t, int objc, Tcl_Obj *CONST objv[])
{
  const cmd_t *cmd;
  int i;
  
  g_assert (objc > 1);

  cmd = g_type_get_qdata (t, cmd_key);

  if (!cmd) {
    appendResult (interp, g_type_name (t), " has no methods");
    return TCL_ERROR;
  }

  /* FIXME: cleanup reporting of all commands in all the classes on error */

  while (Tcl_GetIndexFromObjStruct (interp, objv[1], (char **) &cmd[0].name,
				    sizeof (cmd_t), "cmd", TCL_EXACT, &i)
	 != TCL_OK) {
    do {
      t = g_type_parent (t);
      
      if (t == G_TYPE_INVALID) {
	return TCL_ERROR;
      }
      
    } while (!(cmd = g_type_get_qdata (t, cmd_key)));
  }
  
  /* clear any error msg */
  Tcl_ResetResult (interp);

  cmd += i;

  return do_cmd_eval (cmd, interp, objc, objv);
}

cval_t *cval_gather (Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],
		     int n_args, const cmd_sig_t *sig, guint skip, ...)
{
  cval_t *arg;
  int i, j;

#ifdef TRACE
  cmd_print (sig, objc, objv);
#endif

  if (objc != n_args) {
    va_list ap;
    GString *usage = g_string_sized_new (16);

    va_start (ap, skip);

    for (i = 2, j = 1; i < n_args; i++, skip >>= 1) {
      if (skip & 1) {
	g_string_append (usage, va_arg (ap, const char *));
	g_string_append_c (usage, ' ');	
      } else {
	g_assert (j < sig->len);
	
	g_string_append (usage, g_type_name (sig->type[j++]));
	g_string_append_c (usage, ' ');	
      }
    }

    g_string_truncate (usage, usage->len - 1);
    Tcl_WrongNumArgs (interp, 2, objv, usage->str);

    g_string_free (usage, TRUE);
    va_end (ap);
    return 0;
  }

  arg = g_new0 (cval_t, sig->len);

  for (i = 2, j = 1; i < n_args; i++, skip >>= 1) {
    if (!(skip & 1)) {
      g_assert (j < sig->len);

      if (getType (interp, objv[i], sig->type[j], &arg[j]) != TCL_OK) {
	cval_free (arg, sig->type, j);
	return 0;
      }
      
      j++;
    }
  }

  return arg;
}

int cmd_cmd (ClientData clientData, Tcl_Interp *interp, int objc,
		Tcl_Obj *CONST objv[])
{
  int i;
  GType type;
  cmd_t *cmd;
  static const char *cmds[] = {
    "invoke_type", "invoke_object", "names", "args", (char *) 0
  };
  
  enum {
    CMD_INVOKE_TYPE, CMD_INVOKE_OBJECT, CMD_NAMES, CMD_ARGS
  };

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "command option arg...");
    return TCL_ERROR;
  }
  
  if (Tcl_GetIndexFromObj (interp, objv[1], cmds, "cmd", TCL_EXACT, &i)
      != TCL_OK) {
    return TCL_ERROR;
  }

  switch (i) {
  case CMD_INVOKE_TYPE: {
    Tcl_Obj *obj, **obj_arg;

    if (objc != 5) {
      Tcl_WrongNumArgs (interp, 2, objv, "type cmd {arg}");
      return TCL_ERROR;
    }
    
    if (Tcl_GetIntFromObj (interp, objv[2], &i) != TCL_OK) {
      return TCL_ERROR;
    }
  
    type = i;

    /* arg list must match as if had been called 
     * "g::object set $o name val" and not as
     * "g::cmd invoke_type [g::handle get_type $o] set {$o name val}"
     */

    obj = Tcl_NewListObj (0, 0);
    Tcl_ListObjAppendElement (interp, obj, Tcl_NewObj ());
    Tcl_ListObjAppendElement (interp, obj, objv[3]);
    if (Tcl_ListObjAppendList (interp, obj, objv[4]) != TCL_OK) {
      return TCL_OK;
    }
    Tcl_ListObjGetElements (interp, obj, &i, &obj_arg);

    return cmd_invoke (interp, type, i, obj_arg);
  }

  case CMD_INVOKE_OBJECT: {
    Tcl_Obj *obj, **obj_arg;
    
    void *handle;

    if (objc != 5) {
      Tcl_WrongNumArgs (interp, 2, objv, "GObject cmd {arg}");
      return TCL_ERROR;
    }

    if (getHandle (interp, objv[2], G_TYPE_OBJECT, &handle) != TCL_OK) {
      return TCL_ERROR;
    }

    type = G_OBJECT_TYPE (handle);

    /* arg list must match as if had been called 
     * "g::object set $o name val" and not as
     * "g::cmd invoke_object $o set {name val}"
     */

    obj = Tcl_NewListObj (0, 0);
    Tcl_ListObjAppendElement (interp, obj, Tcl_NewObj ());
    Tcl_ListObjAppendElement (interp, obj, objv[3]);
    Tcl_ListObjAppendElement (interp, obj, objv[2]);
    if (Tcl_ListObjAppendList (interp, obj, objv[4]) != TCL_OK) {
      return TCL_OK;
    }
    Tcl_ListObjGetElements (interp, obj, &i, &obj_arg);

    return cmd_invoke (interp, type, i, obj_arg);
  }

  case CMD_NAMES: {
    Tcl_Obj *obj;

    if (objc != 3) {
      Tcl_WrongNumArgs (interp, 2, objv, "type");
      return TCL_ERROR;
    }
   
    if (Tcl_GetIntFromObj (interp, objv[2], &i) != TCL_OK) {
      return TCL_ERROR;
    }
  
    type = i;

    obj = Tcl_NewListObj (0, 0);
    cmd = g_type_get_qdata (type, cmd_key);
    
    if (cmd) {
      for (i = 0; cmd[i].name; i++) {
	Tcl_ListObjAppendElement (interp, obj,
				  Tcl_NewStringObj (cmd[i].name, -1));
      }
    }

    Tcl_SetObjResult (interp, obj);
    break;
  }

  case CMD_ARGS: {
    if (objc != 4) {
      Tcl_WrongNumArgs (interp, 2, objv, "type cmd");
      return TCL_ERROR;
    }

    if (Tcl_GetIntFromObj (interp, objv[2], &i) != TCL_OK) {
      return TCL_ERROR;
    }
  
    type = i;

    cmd = g_type_get_qdata (type, cmd_key);

    if (cmd) {
      Tcl_Obj *obj;
      const char *s;

      if (Tcl_GetIndexFromObjStruct (interp, objv[3], (char **) &cmd[0].name,
				    sizeof (cmd_t), "cmd", TCL_EXACT, &i)
	 != TCL_OK) {
	  return TCL_ERROR;
      }

      cmd += i;

      obj = Tcl_NewListObj (0, 0);

      if (cmd->sig->wrap) {
	for (i = 0; i < cmd->sig->len; i++) {
	  s = g_type_name (cmd->sig->type[i]);
	  Tcl_ListObjAppendElement (interp, obj, Tcl_NewStringObj (s, -1));
	}
      }

      Tcl_SetObjResult (interp, obj);
    }

    break;
  }
  
  default:
    g_assert_not_reached ();
  }
  
  return TCL_OK;
}
