/**!
 * vte.notes:
 *
 * To set colors see, &{vte::terminal set_colors}.
 **/


static int get_gdk_color (Tcl_Interp *interp, Tcl_Obj *obj, GdkColor *color)
{
  Tcl_Obj **obj2;
  int len, val;
  
  if (Tcl_ListObjGetElements (interp, obj, &len, &obj2) != TCL_OK) {
    return TCL_ERROR;
  }

  if (len != 3) {
    appendResult (interp, "bad color \"",
		  Tcl_GetString (obj),
		  "\" color must be list of form: {red green blue}", 
		  (char *) 0);
    return TCL_ERROR;
  }

  if (Tcl_GetIntFromObj (interp, obj2[0], &val) != TCL_OK) {
    return TCL_ERROR;
  }

  color->red = (guint16) val;

  if (Tcl_GetIntFromObj (interp, obj2[1], &val) != TCL_OK) {
    return TCL_ERROR;
  }

  color->green = (guint16) val;

  if (Tcl_GetIntFromObj (interp, obj2[2], &val) != TCL_OK) {
    return TCL_ERROR;
  }

  color->blue = (guint16) val;

  return TCL_OK;
}

int get_gdk_color_array (Tcl_Interp *interp, Tcl_Obj *obj, GdkColor **palette,
			 int *len)
{
  Tcl_Obj **obj2;
  int i;

  if (Tcl_ListObjGetElements (interp, obj, len, &obj2) != TCL_OK) {
    return TCL_ERROR;
  }

  *palette = g_new (GdkColor, *len);

  for (i = 0; i < *len; i++) {
    if (get_gdk_color (interp, obj2[i], (*palette) + i) != TCL_OK) {
      g_free (*palette);
      return TCL_ERROR;
    }
  }

  return TCL_OK;
}


/**!
 * vte.stub.func.vte::terminal set_colors:
 * vte::terminal set_colors
 * vte_terminal_set_colors
 *
 *> Usage
 *   @name          %VteTerminal*
 *   @name          %foreground - {red green blue}
 *   @name          %background - {red green blue}
 *   @name          %palette    - {{red green blue} ...}
 *   Return value   %void
 *
 * Wraps %vte_terminal_set_colors. As per the C function
 * the palette must be have 8, 16, or 24 colors.
 **/

static int cmd_vte_terminal_set_colors (Tcl_Interp *interp, int objc,
					Tcl_Obj *CONST objv[],
					const cmd_sig_t *sig)
{
  cval_t *arg;
  GdkColor fg, bg, *palette;
  int palette_len;

  /* VTE_TYPE_TERMINAL */
  arg = cval_gather (interp, objc, objv, 6, sig, 0XE,
		     "foreground", "background", "palette");
                                                                               
  if (!arg) {
    return TCL_ERROR;
  }
   

  if (get_gdk_color (interp, objv[3], &fg) != TCL_OK) {
    return TCL_ERROR;
  }

  if (get_gdk_color (interp, objv[4], &bg) != TCL_OK) {
    return TCL_ERROR;
  }

  if (get_gdk_color_array (interp, objv[5], &palette, &palette_len)
      != TCL_OK) {
    return TCL_ERROR;
  }
   
  if (palette_len != 8 && palette_len != 16 && palette_len != 24) {
    appendResult (interp, "palette must have 8, 16, or 24 colors",
		  (char *) 0);
    g_free (palette);
    return TCL_ERROR;
  }
  
  vte_terminal_set_colors (arg[1].ptr_val, &fg, &bg, palette, palette_len);
  g_free (palette);

  cval_free (arg, sig->type, sig->len);
  return TCL_OK;
}
